
/*
    Copyright 2000  Li-Cheng (Andy) Tai

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <assert.h>
#include "util.h"
#include "support.h"
#include "ui.h"
#include "callbacks.h"
#include "gtkscrollframe.h"
#include "pixmaps.h"

GdkCursor *busycursor = 0;

const int erase_size = 4;

GtkWidget *widget_get_toplevel_parent(GtkWidget *widget)
{
   GtkWidget *parent, *found_widget;
   found_widget = (GtkWidget*) gtk_object_get_data(GTK_OBJECT(widget), "toplevelparent");
   if (found_widget)
      return found_widget;
   found_widget = widget;
   for (;;)
     {
       if (GTK_IS_MENU (found_widget))
         parent = gtk_menu_get_attach_widget (GTK_MENU (found_widget));
       else
         parent = found_widget->parent;
       if (parent == NULL)
         break;
       found_widget = parent;
     }
   gtk_object_set_data(GTK_OBJECT(widget), "toplevelparent", (gpointer) found_widget);
   return found_widget;
}

image_buf *widget_get_image(GtkWidget *widget)
{
   image_buf *ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(widget));
   GtkWidget *parent;
   
   if (!ibuf)
   {
      parent = widget_get_toplevel_parent(widget);
      ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(parent));
      gtk_object_set_user_data(GTK_OBJECT(widget), ibuf);
   }
   
   
   return ibuf;
}         

GtkWidget *create_scroll_frame_widget(void)
{
   GtkWidget *scrolledwindow1; 
   scrolledwindow1 = gtk_scroll_frame_new (NULL, NULL);
   return scrolledwindow1;
}

GtkWidget *create_scroll_frame_widget_content(GtkWidget *scrolledwindow1, GtkWidget *mainwindow)
{
   GtkWidget *viewport1, *drawingarea;
   
   gtk_scroll_frame_set_policy (GTK_SCROLL_FRAME (scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

   viewport1 = gtk_viewport_new (NULL, NULL);
   gtk_widget_set_name (viewport1, "viewport1");
   gtk_widget_ref (viewport1);
   gtk_object_set_data_full (GTK_OBJECT (mainwindow), "viewport1", viewport1,
                             (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (viewport1);
   gtk_container_add (GTK_CONTAINER (scrolledwindow1), viewport1);

   drawingarea = gtk_drawing_area_new ();
   gtk_widget_set_name (drawingarea, "drawingarea");
   gtk_widget_ref (drawingarea);
   gtk_object_set_data_full (GTK_OBJECT (mainwindow), "drawingarea", drawingarea,
                             (GtkDestroyNotify) gtk_widget_unref);
   gtk_widget_show (drawingarea);
   gtk_container_add (GTK_CONTAINER (viewport1), drawingarea);
   gtk_widget_set_events (drawingarea, GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON2_MOTION_MASK | GDK_BUTTON3_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK | GDK_PROPERTY_CHANGE_MASK | GDK_VISIBILITY_NOTIFY_MASK);
   GTK_WIDGET_SET_FLAGS(drawingarea, GTK_CAN_FOCUS);
   
   gtk_signal_connect (GTK_OBJECT (drawingarea), "configure_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_configure_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "button_press_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_button_press_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "button_release_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_button_release_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "motion_notify_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_motion_notify_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "realize",
                       GTK_SIGNAL_FUNC (on_drawingarea_realize),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "expose_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_expose_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "delete_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_delete_event),
                       NULL);
   gtk_signal_connect (GTK_OBJECT (drawingarea), "key_release_event",
                       GTK_SIGNAL_FUNC (on_drawingarea_key_release_event),
                       NULL);
   return scrolledwindow1;
}

void new_canvas()
{
   GtkWidget *new_canvas_window = create_new_canvas_window ();
   gtk_widget_show (new_canvas_window);
}   

static void open_canvas_callback(GtkFileSelection *widget)
{
   char tmp[4000];
   image_buf *ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(widget));
   strcpy(tmp, gtk_file_selection_get_filename(GTK_FILE_SELECTION(widget)));
   image_buf_load(ibuf, tmp);
   
   
   gtk_widget_destroy(GTK_WIDGET(widget));
}

   
void open_canvas(image_buf *ibuf)
{
   GtkWidget *widget = gtk_file_selection_new("Open image");
   g_assert(widget);
   /* Set as modal */
   gtk_window_set_modal (GTK_WINDOW(widget),TRUE);

   /* And mark it as a transient dialog */
   gtk_window_set_transient_for (GTK_WINDOW(widget), GTK_WINDOW(ibuf->window));

   gtk_object_set_user_data(GTK_OBJECT(widget), ibuf);
   gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->ok_button),
                             "clicked", GTK_SIGNAL_FUNC(open_canvas_callback),
                             GTK_OBJECT(widget));
   gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->cancel_button),
                             "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                             GTK_OBJECT(widget));

   gtk_widget_show(widget);

}
static void save_canvas_callback(GtkFileSelection *widget)
{
   char tmp[4000];
   image_buf *ibuf = (image_buf*) gtk_object_get_user_data(GTK_OBJECT(widget));

   strcpy(tmp, gtk_file_selection_get_filename(GTK_FILE_SELECTION(widget)));
   image_buf_save(ibuf, tmp);
   gtk_widget_destroy(GTK_WIDGET(widget));
}

void save_canvas(image_buf *ibuf)
{
  
   GtkWidget *widget = gtk_file_selection_new("Save image");
   g_assert(widget);
   /* Set as modal */
   gtk_window_set_modal (GTK_WINDOW(widget),TRUE);

   /* And mark it as a transient dialog */
   gtk_window_set_transient_for(GTK_WINDOW(widget), GTK_WINDOW(ibuf->window));
   if (ibuf->name)
      gtk_file_selection_set_filename(GTK_FILE_SELECTION(widget), ibuf->name);

   gtk_object_set_user_data(GTK_OBJECT(widget), ibuf);
   gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->ok_button),
                             "clicked", GTK_SIGNAL_FUNC(save_canvas_callback),
                             GTK_OBJECT(widget));
   gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(widget)->cancel_button),
                             "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                             GTK_OBJECT(widget));

   gtk_widget_show(widget);

}

static void gtk_container_remove_callback(GtkWidget *widget, gpointer data)
{
   gtk_container_remove(GTK_CONTAINER(data), widget);
}

void set_button_pixmap(GtkButton *button, unsigned char **pixmap)
{
   GdkPixmap *gdkpixmap;
   GdkBitmap *mask;
   GtkWidget *gtkpixmap;
   gdkpixmap = gdk_pixmap_create_from_xpm_d(widget_get_toplevel_parent(GTK_WIDGET(button))->window, &mask, NULL, (gchar**) pixmap);
   g_assert(gdkpixmap);
   gtkpixmap = gtk_pixmap_new(gdkpixmap, mask);
   g_assert(gtkpixmap);
   gdk_pixmap_unref (gdkpixmap);
   gdk_pixmap_unref (mask);
   
   gtk_container_foreach(GTK_CONTAINER(button), (GtkCallback) gtk_container_remove_callback, button);
   gtk_container_add(GTK_CONTAINER(button), (gtkpixmap));
   
   gtk_widget_show(gtkpixmap);
   
}

typedef struct toggle_button_info 
{
   const char *name;
   DRAWING_TOOL tool;
} toggle_button_info;
   
const  toggle_button_info toggle_button_infos[] =
{
   {"erase_button", ERASE},
   {"lasso_button", LASSO},
   {"fill_button", FILL},
   {"line_button", LINE},
   {"multiline_button", MULTILINE},
   {"rectangle_button", RECTANGLE},
   {"closed_freehand_button", CLOSED_FREEHAND},
   {"pen_button", PEN},
   {"polselect_button", POLSELECT},
   {"text_button", TEXT},
   {"arc_button", ARC},
   {"curve_button", CURVE},
   {"oval_button", OVAL},
   {"brush_button", BRUSH},
   {0, NONE}
};
void set_drawing_tool(image_buf *ibuf, DRAWING_TOOL tool)
{
   const char *buttonname = toggle_button_infos[tool - ERASE].name;
   GtkWidget *selected = lookup_widget(ibuf->window, buttonname);
   
   g_assert(selected);
   select_toolbar_toggle_button(GTK_TOGGLE_BUTTON(selected), 1);
}



void select_toolbar_toggle_button(GtkToggleButton *selected, int setmode)
{
   const toggle_button_info *t;
   GtkWidget *b;
   GtkWidget *window ;
   int x, y;
   int inbutton = setmode;
   gtk_widget_get_pointer(GTK_WIDGET(selected), &x, &y);
   if ((x >= 0) && (y >= 0) && (x < GTK_WIDGET(selected)->allocation.width) && (y < GTK_WIDGET(selected)->allocation.height))
      inbutton = 1;
      
   if (!gtk_toggle_button_get_active(selected))
   {
      if (!inbutton)
         return;
      
      gtk_toggle_button_set_active(selected, TRUE);
      
   }
   
   window = widget_get_toplevel_parent(GTK_WIDGET(selected));
   g_assert(window); 
   if (inbutton)
   {
      image_buf *ibuf = widget_get_image(window);
      g_assert(ibuf);
      ibuf->current_button = GTK_WIDGET(selected);
   }
   
   for (t = toggle_button_infos; (t->name); t++)
   {
      b = lookup_widget(window, (t->name));
      g_assert(b);
      if (b != GTK_WIDGET(selected))      
      {
         if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(b)))
	 {
	    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), FALSE);
          
	 }        
         g_assert(!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(b)));
      }
      else if (inbutton)
      {
	 image_buf *ibuf = widget_get_image(window);
	 g_assert(ibuf);
	 image_buf_set_tool(ibuf, t->tool);
      }
         
   }
}
      
void StateTimeStep()
{

}      


void image_buf_set_fill(image_buf *ibuf, int filled)
{
   GtkWidget *b = lookup_widget(ibuf->window, "filled_button");
   int old = ibuf->filled;
   assert(b);
   ibuf->filled = filled;
   if (filled != old)
   {
      if (filled)
         set_button_pixmap(GTK_BUTTON(b),  (unsigned char * *) filled_xpm);
      else
         set_button_pixmap(GTK_BUTTON(b),  (unsigned char * *) unfilled_xpm);
   }
   
}

static void clear_points(image_buf *ibuf)
{
   ibuf->num_pts = 0;

   memset(ibuf->pts, 0, sizeof(ibuf->pts));
}


void image_buf_set_tool(image_buf *ibuf, DRAWING_TOOL tool)
{

   switch(ibuf->current_tool) /* final mode cleanup */
   {
   case TEXT:
      if ((ibuf->llx > INT_MIN) && (ibuf->lly > INT_MIN) && strlen(ibuf->textbuf))
      {
	 gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
	 gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
      }
      ibuf->llx = INT_MIN;
      ibuf->lly = INT_MIN;
      break;
   
   default:
      break;
   }
   
   ibuf->current_tool = tool;
   switch(tool) /* new mode initialization */
   {
   case MULTILINE:
   
   case ERASE:
   case LINE:
   case ARC:
   case RECTANGLE:
   case OVAL:
   
      image_buf_set_cursor(ibuf, gdk_cursor_new(GDK_CROSSHAIR));
      break;
   
   case CLOSED_FREEHAND:
   case PEN:
      image_buf_set_cursor(ibuf, gdk_cursor_new(GDK_PENCIL));
      break;
    
   case TEXT:
      image_buf_set_cursor(ibuf, gdk_cursor_new(GDK_XTERM));
      strcpy(ibuf->textbuf, "");
      break;
      
   default:
      image_buf_set_cursor(ibuf, gdk_cursor_new(GDK_X_CURSOR));
      break;
   }
   ibuf->llx = INT_MIN;
   ibuf->lly = INT_MIN;
   ibuf->lx = INT_MIN;
   ibuf->ly = INT_MIN;
   ibuf->mx = INT_MIN;
   ibuf->my = INT_MIN;
   clear_points(ibuf);
}
   

static void insert_point(image_buf *ibuf, int x, int y)
{
   ibuf->pts[ibuf->num_pts].x = x;
   ibuf->pts[ibuf->num_pts].y = y;
   ibuf->num_pts++;
}

static void draw_rectangle(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, gint x2, gint y2)
{
   int ox, w, oy, h;
   if (x1 > x2)
   {
      ox = x2;
      w = x1 - x2;
   }
   else
   {
      ox = x1;
      w = x2 - x1;
   }
   
   if (y1 > y2)
   {
      oy = y2;
      h = y1 - y2;
   }
   else
   {
      oy = y1;
      h = y2 - y1;
   }
   if (filled)
      w++, h++;
   gdk_draw_rectangle(d, gc, filled, ox, oy, w, h);
}

static void draw_oval(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, gint x2, gint y2)
{
   int ox, w, oy, h;
   if (x1 > x2)
   {
      ox = x2;
      w = x1 - x2;
   }
   else
   {
      ox = x1;
      w = x2 - x1;
   }
   
   if (y1 > y2)
   {
      oy = y2;
      h = y1 - y2;
   }
   else
   {
      oy = y1;
      h = y2 - y1;
   }
   if (filled)
      w++, h++;
   gdk_draw_arc(d, gc, filled, ox, oy, w, h, 0, 360 * 64);
}


static void draw_arc(GdkDrawable *d, GdkGC *gc, gint filled, gint x1, gint y1, gint x2, gint y2)
{
   int ox, w, oy, h;
   int angle1, angle2;
   
   
   if (x1 < x2)
   {
      w = (x2 - x1) * 2;
     
      if (y1 < y2)
      {
         h = (y2 - y1) * 2;  
         ox = x2 - w;
	 oy = y1;
         angle1 = 90;
	 angle2 = -90;
      }
      else
      {
         ox = x1;
         h = (y1 - y2) * 2;  
	 oy = y2;
         angle1 = 180;
	 angle2 = -90;
      }
   }
   else /* x1 > x2 */
   {
      w = (x1 - x2) * 2;
      
      if (y1 < y2)
      {
         h = (y2 - y1) * 2;
	 ox = x1 - w;
	 oy = y2 - h;
         angle1 = 0;
	 angle2 = -90;
      }
      else
      {
         h = (y1 - y2) * 2;
	 oy = y1 - h;
	 ox = x2;
         angle1 = -90;
	 angle2 = -90;
      }
   }
      
   if (filled)
      w++, h++;
      
   gdk_draw_arc(d, gc, filled, ox, oy, w, h, angle1 * 64, angle2 * 64);
}

void handle_button_press(image_buf *ibuf, int x, int y)
{
   DRAWING_TOOL tool = ibuf->current_tool;
   GdkGCValues gcvalues;
   GdkRectangle rect, clientrect ;
      
   rect.x = x;
   rect.y = y;
   rect.width = 10;
   rect.height = 10;
   clientrect.x = clientrect.y = 0;
   clientrect.width = image_buf_width(ibuf);
   clientrect.height = image_buf_height(ibuf);
   
   gdk_gc_get_values(ibuf->gc, &gcvalues);
   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
   
   insert_point(ibuf, x, y);
   switch (tool)
   {
   case CLOSED_FREEHAND:
   case PEN:
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, GDK_CAP_ROUND, GDK_JOIN_ROUND);
      gdk_draw_line(ibuf->pixmap, ibuf->gc, x, y, x, y);
      gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, x, y, x, y);
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
      //gtk_widget_draw(ibuf->drawing_area, &rect);
      break;
   
   case MULTILINE:
      if ((ibuf->llx > INT_MIN) && (ibuf->lly > INT_MIN))
      {
	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, x, y, ibuf->llx, ibuf->lly);
	 gdk_draw_line(ibuf->pixmap, ibuf->gc, x, y, ibuf->llx, ibuf->lly);
      }      
      
   case LINE:
      break;
     
   case OVAL:
      break;
   
   case ARC:
      break;
      
   case RECTANGLE:
      break;	 
	
   case TEXT:
      break;
   	
   case ERASE:
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
      draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
      break;
      
   case FILL:
      image_buf_flood_fill(ibuf, x, y, &(gcvalues.foreground));	
      gtk_widget_draw(ibuf->drawing_area, NULL);
      break;
      	 
   default:
      gdk_beep();
   }
   ibuf->lx = x;
   ibuf->ly = y;
   ibuf->mx = INT_MIN;
   ibuf->my = INT_MIN;
}   

void handle_button_move(image_buf *ibuf, int x, int y)
{
   DRAWING_TOOL tool = ibuf->current_tool;
   GdkGCValues gcvalues;
   GdkRectangle rect, clientrect ;
   
   if ((ibuf->lx == INT_MIN) || (ibuf->ly == INT_MIN))
      return;
      
   rect.x = x;
   rect.y = y;
   rect.width = 10;
   rect.height = 10;
   clientrect.x = clientrect.y = 0;
   clientrect.width = image_buf_width(ibuf);
   clientrect.height = image_buf_height(ibuf);
   
   gdk_gc_get_values(ibuf->gc, &gcvalues);
   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
   
   insert_point(ibuf, x, y);
   
   switch (tool)
   {
   case CLOSED_FREEHAND:
   case PEN:
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, GDK_CAP_ROUND, GDK_JOIN_ROUND);
       
      if ((ibuf->mx == INT_MIN) && (ibuf->my == INT_MIN))
      {
	 gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
      }
      else
      {
	 gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->mx, ibuf->my, x, y);
	 gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->mx, ibuf->my, x, y);
      }
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
      //gtk_widget_draw(ibuf->drawing_area, &rect);
      break;
      
   case MULTILINE:
   case LINE:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
      {
         gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      }
      gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
      gdk_gc_set_function(ibuf->gc, gcvalues.function);  
      break;
      
   case OVAL:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
      {
         draw_oval(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      }
      draw_oval(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, x, y);
      gdk_gc_set_function(ibuf->gc, gcvalues.function);  
      break;
   
   case ARC:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
      {
         draw_arc(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      }
      draw_arc(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, x, y);
      gdk_gc_set_function(ibuf->gc, gcvalues.function);  
      break;
   
   case RECTANGLE:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
      {
         draw_rectangle(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      }
      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, x, y);
      gdk_gc_set_function(ibuf->gc, gcvalues.function);  
      
      break;	 
   
   case ERASE:
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
      draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
      break;
       
   default:
      
   }
   ibuf->mx = x;
   ibuf->my = y;
}   

void handle_button_release(image_buf *ibuf, int x, int y)
{
   DRAWING_TOOL tool = ibuf->current_tool;
   GdkGCValues gcvalues;
   GdkRectangle rect, clientrect ;
   
   if ((ibuf->lx == INT_MIN) || (ibuf->ly == INT_MIN))
      return;
      
   rect.x = x;
   rect.y = y;
   rect.width = 10;
   rect.height = 10;
   clientrect.x = clientrect.y = 0;
   clientrect.width = image_buf_width(ibuf);
   clientrect.height = image_buf_height(ibuf);
   
   gdk_gc_get_values(ibuf->gc, &gcvalues);
   gdk_gc_set_clip_origin(ibuf->gc, 0, 0);
   gdk_gc_set_clip_rectangle(ibuf->gc, &clientrect);
   insert_point(ibuf, x, y);
   
   switch (tool)
   {
   case CLOSED_FREEHAND:
   case PEN:
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, GDK_CAP_ROUND, GDK_JOIN_ROUND);
      if ((ibuf->mx == INT_MIN) && (ibuf->my == INT_MIN))
      {
         if (tool == CLOSED_FREEHAND)
	 {
	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
	 
	 }
	 else
	 {
	    gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
	    gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
	 }
      }
      else
      {
         if (tool == CLOSED_FREEHAND)
	 {
	    gdk_draw_polygon(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
	    gdk_draw_polygon(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->pts, ibuf->num_pts);
	 }
	 else
	 {
	    gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->mx, ibuf->my, x, y);
	    gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->mx, ibuf->my, x, y);
	 }
      }
      gdk_gc_set_line_attributes(ibuf->gc, gcvalues.line_width, gcvalues.line_style, gcvalues.cap_style, gcvalues.join_style);
      //gtk_widget_draw(ibuf->drawing_area, &rect);
      break;
      
   case MULTILINE:
   case LINE:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
         gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      
      gdk_gc_set_function(ibuf->gc, gcvalues.function);   
      gdk_draw_line(ibuf->pixmap, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
      gdk_draw_line(ibuf->drawing_area->window, ibuf->gc, ibuf->lx, ibuf->ly, x, y);
      break;
      
   case OVAL:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
         draw_oval(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      
      gdk_gc_set_function(ibuf->gc, gcvalues.function);   
      draw_oval(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      draw_oval(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      break;
   
   case ARC:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
         draw_arc(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      
      gdk_gc_set_function(ibuf->gc, gcvalues.function);   
      draw_arc(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      draw_arc(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      break;
      
      
   case RECTANGLE:
      gdk_gc_set_function(ibuf->gc, GDK_XOR);
      if ((ibuf->mx > INT_MIN) && (ibuf->my > INT_MIN))
         draw_rectangle(ibuf->drawing_area->window, ibuf->gc, FALSE, ibuf->lx, ibuf->ly, ibuf->mx, ibuf->my);
      
      gdk_gc_set_function(ibuf->gc, gcvalues.function);   
      draw_rectangle(ibuf->pixmap, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, ibuf->filled, ibuf->lx, ibuf->ly, x, y);
      break;	 
   
  
   case ERASE:
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.background));
      draw_rectangle(ibuf->pixmap, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      draw_rectangle(ibuf->drawing_area->window, ibuf->gc, TRUE, x - erase_size, y - erase_size, x + erase_size, y + erase_size);
      
      gdk_gc_set_foreground(ibuf->gc, &(gcvalues.foreground));
      break;
         
   case TEXT:
      gtk_widget_grab_focus(ibuf->drawing_area);
      break;
      
   default:
      
   }
   ibuf->lx = INT_MIN;
   ibuf->ly = INT_MIN;
   ibuf->mx = INT_MIN;
   ibuf->my = INT_MIN;
   if (tool == MULTILINE)
   {
      ibuf->llx = x;
      ibuf->lly = y;
   }
   
   if (tool == TEXT)
   {
      ibuf->llx = x;
      ibuf->lly = y;
   }
      
   clear_points(ibuf);
}   


void handle_key_release(image_buf *ibuf, GdkEventKey *keyevent)
{
   GdkGCValues gcvalues;
   
   int length;
   char tmp[MAX_TEXT_LENGTH];
   if ((ibuf->llx == INT_MIN) || (ibuf->lly == INT_MIN) || (ibuf->current_tool != TEXT))
   {
      gdk_beep();
      return;
   }
   
   
   gdk_gc_get_values(ibuf->gc, &gcvalues);
   strcpy(tmp, ibuf->textbuf);
   length = strlen(ibuf->textbuf);
   if ((keyevent->keyval == GDK_BackSpace) || (keyevent->keyval == GDK_Delete))
   {
      if (length)
         ibuf->textbuf[length - 1] = 0;
   }
   else if (keyevent->string)
      strcat(ibuf->textbuf, keyevent->string);
   else if (keyevent->keyval >= GDK_space)
   {
      sprintf(ibuf->textbuf, "%s%c", tmp, keyevent->keyval);
   }
   
   gdk_gc_set_function(ibuf->gc, GDK_XOR);
      
   if (strlen(tmp))
   {
      gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, tmp);
      gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, tmp);
   }
    
   if (strlen(ibuf->textbuf))
   {
      gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
      gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
   }
   
   
   gdk_gc_set_function(ibuf->gc, gcvalues.function);   

   if (keyevent->keyval == GDK_Return)
   {
      if (strlen(ibuf->textbuf))
      {
	 gdk_draw_string(ibuf->drawing_area->window, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
	 gdk_draw_string(ibuf->pixmap, ibuf->font, ibuf->gc, ibuf->llx, ibuf->lly, ibuf->textbuf);
      }
      strcpy(ibuf->textbuf, "");
      ibuf->llx = INT_MIN;
      ibuf->lly = INT_MIN;
   }
   
}


void GdkColor_to_rgb(const GdkColor *color, unsigned char *r, unsigned char *g, unsigned char *b)
{
   
   GdkColormap *cmap = gdk_rgb_get_cmap();
   GdkVisual *visual = gdk_rgb_get_visual();
   GdkColor *c;
   switch(visual->type)
   {
   case GDK_VISUAL_STATIC_COLOR:
   case GDK_VISUAL_PSEUDO_COLOR:
      c = cmap->colors + color->pixel;
  
      *r = c->red >> 8;
      *g = c->green >> 8;
      *b = c->blue >> 8;
      break;
   case GDK_VISUAL_TRUE_COLOR:
   case GDK_VISUAL_DIRECT_COLOR:
      *r = (color->pixel & visual->red_mask) >> visual->red_shift;
      *g = (color->pixel & visual->green_mask) >> visual->green_shift;
      *b = (color->pixel & visual->blue_mask) >> visual->blue_shift;
      break;
   
   default:
      assert(0);
      break;
   }
 }
      
#include "image_processing.h"

void spreadimage(image_buf * input, image_buf *output)
{
   gdk_pixbuf_copy_area(input->rgbbuf, 0, 0, image_buf_width(input), image_buf_height(input), output->rgbbuf, 0, 0);
   ImageSpread(output);
}
